# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import inspect
from typing import Optional, Union, Callable
from base64 import b64encode
from textwrap import dedent
from servicex.query_core import QueryStringGenerator
import sys

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class PythonFunction(QueryStringGenerator):
    yaml_tag = "!PythonFunction"
    default_codegen = "python"

    def __init__(self, python_function: Optional[Union[str, Callable]] = None):
        self.python_function: Optional[Union[str, Callable]] = python_function

    def with_uproot_function(self, f: Union[str, Callable]) -> Self:
        self.python_function = f
        return self

    def generate_selection_string(self) -> str:
        if not self.python_function:
            raise ValueError(
                "You must provide a python function using with_uproot_function"
            )

        if isinstance(self.python_function, str):
            return b64encode(dedent(self.python_function).encode("utf-8")).decode(
                "utf-8"
            )
        else:
            return b64encode(
                dedent(inspect.getsource(self.python_function)).encode("utf-8")
            ).decode("utf-8")

    @classmethod
    def from_yaml(cls, _, node):
        code = node.value
        try:
            exec(code)
        except SyntaxError as e:
            raise SyntaxError(f"Syntax error {e} interpreting\n{code}")
        q = PythonFunction(code)
        return q



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import tempfile
import time
from unittest.mock import patch

import httpx
import pytest
from aiohttp import ContentTypeError
from pytest_asyncio import fixture

from servicex.models import TransformRequest, ResultDestination, ResultFormat
from servicex.servicex_adapter import ServiceXAdapter, AuthorizationError


@fixture
def servicex():
    return ServiceXAdapter("https://servicex.org")


def test_result_formats():
    """
    This test is just to make sure the enum string representations match the values
    expected by the service. Compare this to the json parser in
    servicex.resources.transformation.submit.SubmitTransformationRequest.make_api
    """
    assert ResultFormat.parquet == "parquet"
    assert ResultFormat.root_ttree == "root-file"


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transforms(mock_get, servicex, transform_status_response):
    mock_get.return_value.__aenter__.return_value.json.return_value = (
        transform_status_response
    )
    mock_get.return_value.__aenter__.return_value.status = 200
    t = await servicex.get_transforms()
    assert len(t) == 1
    assert t[0].request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
    mock_get.assert_called_with(
        url="https://servicex.org/servicex/transformation", headers={}
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transforms_error(mock_get, servicex, transform_status_response):
    mock_get.return_value.__aenter__.return_value.json.return_value = {
        "message": "error_message"
    }
    mock_get.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.get_transforms()
        assert (
            "ServiceX WebAPI Error during transformation submission: 500 - error_message"
            == str(err.value)
        )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transforms_auth_error(mock_get, servicex):
    with pytest.raises(AuthorizationError) as err:
        mock_get.return_value.__aenter__.return_value.status = 401
        await servicex.get_transforms()
        assert "Not authorized to access serviceX at" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.jwt.decode")
async def test_get_transforms_wlcg_bearer_token(
    decode, servicex, transform_status_response
):
    token_file = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
    token_file.write(
        """"
    eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c
    """
    )
    token_file.close()

    os.environ["BEARER_TOKEN_FILE"] = token_file.name

    # Try with an expired token
    with pytest.raises(AuthorizationError) as err:
        decode.return_value = {"exp": 0.0}
        await servicex.get_transforms()
        assert "ServiceX access token request rejected:" in str(err.value)

    os.remove(token_file.name)
    del os.environ["BEARER_TOKEN_FILE"]


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.post")
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transforms_with_refresh(get, post, transform_status_response):
    servicex = ServiceXAdapter(url="https://servicex.org", refresh_token="refrescas")
    post.return_value.__aenter__.return_value.json.return_value = {
        "access_token": "luckycharms"
    }
    post.return_value.__aenter__.return_value.status = 200
    get.return_value.__aenter__.return_value.json.return_value = (
        transform_status_response
    )
    get.return_value.__aenter__.return_value.status = 200
    await servicex.get_transforms()

    post.assert_called_with(
        "https://servicex.org/token/refresh",
        headers={"Authorization": "Bearer refrescas"},
        json=None,
    )

    get.assert_called_with(
        url="https://servicex.org/servicex/transformation",
        headers={"Authorization": "Bearer luckycharms"},
    )


@patch("servicex.servicex_adapter.httpx.Client.get")
def test_get_codegens(get, servicex):
    get.return_value = httpx.Response(
        200, json={"uproot": "http://uproot-codegen", "xaod": "http://xaod-codegen"}
    )
    c = servicex.get_code_generators()
    assert len(c) == 2
    assert c["uproot"] == "http://uproot-codegen"


@patch("servicex.servicex_adapter.httpx.Client.get")
def test_get_codegens_error(get, servicex):
    get.return_value = httpx.Response(403)
    with pytest.raises(AuthorizationError) as err:
        servicex.get_code_generators()
        assert "Not authorized to access serviceX at" in str(err.value)


@fixture
def dataset():
    return {
        "id": 123,
        "name": "rucio://user.mtost:user.mtost.700349.Sh.DAOD_PHYS.e8351_s3681_r13144_r13146_p6026.Jul13_less_jet_and_new_GN?files=7",  # NOQA: E501
        "did_finder": "rucio",
        "n_files": 7,
        "size": 1359895862,
        "events": 0,
        "last_used": "2024-11-12T01:59:19.161655Z",
        "last_updated": "1969-12-31T18:00:00.000000Z",
        "lookup_status": "complete",
        "is_stale": False,
        "files": [
            {
                "id": 12,
                "adler32": "62c594d4",
                "file_size": 34831129,
                "file_events": 0,
                "paths": "https://xenia.nevis.columbia.edu:1094/atlas/dq2/rucio/user/mtost/06/a1/user.mtost.40294033._000002.less_jet_and_new_GN.root",  # NOQA: E501
            }
        ],
    }


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_get_datasets(get, servicex, dataset):
    get.return_value.__aenter__.return_value.json.return_value = {"datasets": [dataset]}
    get.return_value.__aenter__.return_value.status = 200

    c = await servicex.get_datasets()
    assert len(c) == 1
    assert c[0].id == 123
    get.assert_called_with(
        url="https://servicex.org/servicex/datasets", params={}, headers={}
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_get_datasets_show_deleted(get, servicex, dataset):
    get.return_value.__aenter__.return_value.json.return_value = {"datasets": [dataset]}
    get.return_value.__aenter__.return_value.status = 200
    c = await servicex.get_datasets(show_deleted=True)
    assert len(c) == 1
    assert c[0].id == 123
    get.assert_called_with(
        url="https://servicex.org/servicex/datasets",
        params={"show-deleted": True},
        headers={},
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_get_datasets_auth_error(get, servicex):
    get.return_value.__aenter__.return_value.status = 403
    with pytest.raises(AuthorizationError) as err:
        await servicex.get_datasets()
    assert "Not authorized to access serviceX at" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_get_dataset(get, servicex, dataset):
    get.return_value.__aenter__.return_value.json.return_value = dataset
    get.return_value.__aenter__.return_value.status = 200
    c = await servicex.get_dataset(123)
    assert c
    assert c.id == 123


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_get_dataset_errors(get, servicex, dataset):
    get.return_value.__aenter__.return_value.status = 403
    with pytest.raises(AuthorizationError) as err:
        await servicex.get_dataset(123)
    assert "Not authorized to access serviceX at" in str(err.value)

    get.return_value.__aenter__.return_value.status = 404
    with pytest.raises(ValueError) as err:
        await servicex.get_dataset(123)
    assert "Dataset 123 not found" in str(err.value)

    get.return_value.__aenter__.return_value.json.side_effect = ContentTypeError(
        None, None
    )
    get.return_value.__aenter__.return_value.text.return_value = "error_message"
    get.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.get_dataset(123)
    assert "Failed to get dataset 123 - error_message" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.delete")
async def test_delete_dataset(delete, servicex):
    delete.return_value.__aenter__.return_value.json.return_value = {
        "dataset-id": 123,
        "stale": True,
    }
    delete.return_value.__aenter__.return_value.status = 200

    r = await servicex.delete_dataset(123)
    delete.assert_called_with(
        url="https://servicex.org/servicex/datasets/123", headers={}
    )
    assert r


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.delete")
async def test_delete_dataset_errors(delete, servicex):
    delete.return_value.__aenter__.return_value.status = 403
    with pytest.raises(AuthorizationError) as err:
        await servicex.delete_dataset(123)
    assert "Not authorized to access serviceX at" in str(err.value)

    delete.return_value.__aenter__.return_value.status = 404
    with pytest.raises(ValueError) as err:
        await servicex.delete_dataset(123)
    assert "Dataset 123 not found" in str(err.value)

    delete.return_value.__aenter__.return_value.json.side_effect = ContentTypeError(
        None, None
    )
    delete.return_value.__aenter__.return_value.text.return_value = "error_message"
    delete.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.delete_dataset(123)
    assert "Failed to delete dataset 123 - error_message" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.delete")
async def test_delete_transform(delete, servicex):
    delete.return_value.__aenter__.return_value.status = 200
    await servicex.delete_transform("123-45-6789")
    delete.assert_called_with(
        url="https://servicex.org/servicex/transformation/123-45-6789", headers={}
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.delete")
async def test_delete_transform_errors(delete, servicex):
    delete.return_value.__aenter__.return_value.status = 403
    with pytest.raises(AuthorizationError) as err:
        await servicex.delete_transform("123-45-6789")
    assert "Not authorized to access serviceX at" in str(err.value)

    delete.return_value.__aenter__.return_value.status = 404
    with pytest.raises(ValueError) as err:
        await servicex.delete_transform("123-45-6789")
    assert "Transform 123-45-6789 not found" in str(err.value)

    delete.return_value.__aenter__.return_value.json.side_effect = ContentTypeError(
        None, None
    )
    delete.return_value.__aenter__.return_value.text.return_value = "error_message"
    delete.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.delete_transform("123-45-6789")
    assert "Failed to delete transform 123-45-6789 - error_message" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_cancel_transform(get, servicex):
    get.return_value.__aenter__.return_value.json.return_value = {
        "message": "Canceled transformation request 123"
    }
    get.return_value.__aenter__.return_value.status = 200

    await servicex.cancel_transform(123)
    get.assert_called_with(
        url="https://servicex.org/servicex/transformation/123/cancel", headers={}
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.ClientSession.get")
async def test_cancel_transform_errors(get, servicex):
    get.return_value.__aenter__.return_value.status = 403
    with pytest.raises(AuthorizationError) as err:
        await servicex.cancel_transform(123)
    assert "Not authorized to access serviceX at" in str(err.value)

    get.return_value.__aenter__.return_value.status = 404
    with pytest.raises(ValueError) as err:
        await servicex.cancel_transform(123)
    assert "Transform 123 not found" in str(err.value)

    get.return_value.__aenter__.return_value.json.side_effect = ContentTypeError(
        None, None
    )
    get.return_value.__aenter__.return_value.text.return_value = "error_message"
    get.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.cancel_transform(123)
    assert "Failed to cancel transform 123 - error_message" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.post")
async def test_submit(post, servicex):
    post.return_value.__aenter__.return_value.json.return_value = {
        "request_id": "123-456-789"
    }
    post.return_value.__aenter__.return_value.status = 200
    request = TransformRequest(
        title="Test submission",
        did="rucio://foo.bar",
        selection="(call EventDataset)",
        codegen="uproot",
        result_destination=ResultDestination.object_store,
        result_format=ResultFormat.parquet,
    )
    result = await servicex.submit_transform(request)
    assert result == "123-456-789"


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.post")
async def test_submit_errors(post, servicex):
    post.return_value.__aenter__.return_value.status = 401
    request = TransformRequest(
        title="Test submission",
        did="rucio://foo.bar",
        selection="(call EventDataset)",
        codegen="uproot",
        result_destination=ResultDestination.object_store,
        result_format=ResultFormat.parquet,
    )
    with pytest.raises(AuthorizationError) as err:
        await servicex.submit_transform(request)
    assert "Not authorized to access serviceX at" in str(err.value)

    post.return_value.__aenter__.return_value.json.side_effect = ContentTypeError(
        None, None
    )
    post.return_value.__aenter__.return_value.text.return_value = "error_message"
    post.return_value.__aenter__.return_value.status = 500
    with pytest.raises(RuntimeError) as err:
        await servicex.submit_transform(request)
    assert (
        "ServiceX WebAPI Error during transformation submission: 500 - error_message"
        == str(err.value)
    )

    post.return_value.__aenter__.return_value.json.reset_mock()
    post.return_value.__aenter__.return_value.json.return_value = {
        "message": "error_message"
    }
    post.return_value.__aenter__.return_value.status = 400
    with pytest.raises(ValueError) as err:
        await servicex.submit_transform(request)
    assert "Invalid transform request: error_message" == str(err.value)

    post.return_value.__aenter__.return_value.json.return_value = {
        "message": "error_message"
    }
    post.return_value.__aenter__.return_value.status = 410
    with pytest.raises(RuntimeError) as err:
        await servicex.submit_transform(request)
    assert (
        "ServiceX WebAPI Error during transformation submission: 410 - error_message"
        == str(err.value)
    )


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transform_status(get, servicex, transform_status_response):
    get.return_value.__aenter__.return_value.json.return_value = (
        transform_status_response["requests"][0]
    )  # NOQA: E501
    get.return_value.__aenter__.return_value.status = 200
    result = await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert result.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_transform_status_errors(get, servicex):
    with pytest.raises(AuthorizationError) as err:
        get.return_value.__aenter__.return_value.status = 401
        await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert "Not authorized to access serviceX at " in str(err.value)

    with pytest.raises(ValueError) as err:
        get.return_value.__aenter__.return_value.status = 404
        await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert "Transform ID b8c508d0-ccf2-4deb-a1f7-65c839eebabf not found" == str(
        err.value
    )

    with pytest.raises(RuntimeError) as err:
        get.return_value.__aenter__.return_value.status = 500

        async def patch_json():
            return {"message": "fifteen"}

        get.return_value.__aenter__.return_value.json = patch_json
        await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert "ServiceX WebAPI Error during transformation" in str(err.value)


@pytest.mark.asyncio
@patch("servicex.servicex_adapter.TransformStatus", side_effect=RuntimeError)
@patch("servicex.servicex_adapter.RetryClient.get")
async def test_get_tranform_status_retry_error(
    get, mock_transform_status, servicex, transform_status_response
):
    with pytest.raises(RuntimeError) as err:
        get.return_value.__aenter__.return_value.json.return_value = (
            transform_status_response["requests"][0]
        )  # NOQA: E501
        get.return_value.__aenter__.return_value.status = 200
        await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert "ServiceX WebAPI Error while getting transform status:" in str(err.value)


@pytest.mark.asyncio
async def test_get_authorization(servicex):
    servicex.token = "token"
    servicex.refresh_token = "refresh"
    with patch("google.auth.jwt.decode", return_value={"exp": time.time() + 90}):
        r = await servicex._get_authorization()
        assert r.get("Authorization") == "Bearer token"

    with patch(
        "servicex.servicex_adapter.ServiceXAdapter._get_token", return_value="token"
    ) as get_token:
        with patch("google.auth.jwt.decode", return_value={"exp": time.time() - 90}):
            r = await servicex._get_authorization()
            get_token.assert_called_once()



# Copyright (c) 2022-2025, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from unittest.mock import AsyncMock
import urllib.parse

import pytest
from pytest_asyncio import fixture

from servicex.minio_adapter import MinioAdapter
from servicex.models import ResultFile

DOWNLOAD_PATCH_COUNTER = 0


def mock_downloader(**args):
    global DOWNLOAD_PATCH_COUNTER
    DOWNLOAD_PATCH_COUNTER += 1
    if DOWNLOAD_PATCH_COUNTER == 1:
        raise Exception("lol")
    elif DOWNLOAD_PATCH_COUNTER == 2:
        open("/tmp/foo/test.txt", "wb").write(b"\x01" * 5)
    elif DOWNLOAD_PATCH_COUNTER == 3:
        open("/tmp/foo/test.txt", "wb").write(b"\x01" * 10)


@fixture
def download_patch():
    import aioboto3.s3.inject

    aioboto3.s3.inject.download_file = AsyncMock(side_effect=mock_downloader)
    return aioboto3.s3.inject.download_file


@fixture
def minio_adapter(moto_services, moto_patch_session) -> MinioAdapter:
    urlinfo = urllib.parse.urlparse(moto_services["s3"])
    return MinioAdapter(
        urlinfo.netloc, urlinfo.scheme == "https", "access_key", "secret_key", "bucket"
    )


@fixture
async def populate_bucket(request, minio_adapter):
    async with minio_adapter.minio.client(
        "s3", endpoint_url=minio_adapter.endpoint_host
    ) as s3:
        await s3.create_bucket(Bucket=minio_adapter.bucket)
        await s3.put_object(
            Bucket=minio_adapter.bucket, Key=request.param, Body=b"\x01" * 10
        )
        yield
        await s3.delete_object(Bucket=minio_adapter.bucket, Key=request.param)


@fixture
def session(mocker):
    mock_session = mocker.MagicMock()
    mock_ctor = mocker.patch("servicex.minio_adapter.aiohttp.ClientSession")
    mock_ctor.return_value.__enter__ = mock_session
    return mock_ctor


@pytest.mark.asyncio
async def test_initialize_from_status(completed_status):
    minio = MinioAdapter.for_transform(completed_status)
    assert minio.endpoint_host == "http://minio.org:9000"
    assert minio.minio._session._credentials.access_key == "miniouser"
    assert minio.minio._session._credentials.secret_key == "secret"
    assert minio.bucket == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"


@pytest.mark.parametrize("populate_bucket", ["test.txt"], indirect=True)
@pytest.mark.asyncio
async def test_list_bucket(minio_adapter, populate_bucket):
    files = [ResultFile(filename="test.txt", size=10, extension="txt")]
    result = await minio_adapter.list_bucket()
    assert result == files


@pytest.mark.parametrize("populate_bucket", ["test.txt"], indirect=True)
@pytest.mark.asyncio
async def test_download_file(minio_adapter, populate_bucket):
    result = await minio_adapter.download_file("test.txt", local_dir="/tmp/foo")
    assert str(result).endswith("test.txt")
    assert result.exists()
    assert result.read_bytes() == (b"\x01" * 10)
    result.unlink()  # it should exist, from above ...


@pytest.mark.parametrize("populate_bucket", ["t::est.txt"], indirect=True)
@pytest.mark.asyncio
async def test_download_bad_filename(minio_adapter, populate_bucket):
    result = await minio_adapter.download_file("t::est.txt", local_dir="/tmp/foo")
    assert str(result).endswith("t__est.txt")
    assert result.exists()
    assert result.read_bytes() == (b"\x01" * 10)
    result.unlink()  # it should exist, from above ...


@pytest.mark.parametrize("populate_bucket", ["test.txt"], indirect=True)
@pytest.mark.asyncio
async def test_download_short_filename_no_change(minio_adapter, populate_bucket):
    result = await minio_adapter.download_file(
        "test.txt", local_dir="/tmp/foo", shorten_filename=True
    )
    assert str(result).endswith("test.txt")
    assert result.exists()
    assert result.read_bytes() == (b"\x01" * 10)
    result.unlink()  # it should exist, from above ...


@pytest.mark.parametrize(
    "populate_bucket",
    [
        "test12345678901234567890123456789012345678901234567898012345678901234567890.txt"  # noqa: E501
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_download_short_filename_change(minio_adapter, populate_bucket):
    result = await minio_adapter.download_file(
        "test12345678901234567890123456789012345678901234567898012345678901234567890.txt",
        local_dir="/tmp/foo",
        shorten_filename=True,
    )

    # Some of the filename should be left over still...
    assert str(result).endswith("01234567890.txt")

    # Make sure the length is right
    assert len(result.name) == 60

    assert result.exists()
    assert result.read_bytes() == (b"\x01" * 10)
    result.unlink()  # it should exist, from above ...


@pytest.mark.parametrize("populate_bucket", ["test.txt"], indirect=True)
@pytest.mark.asyncio
async def test_download_repeat(minio_adapter, populate_bucket):
    import asyncio

    print(await minio_adapter.list_bucket())
    result = await minio_adapter.download_file("test.txt", local_dir="/tmp/foo")
    assert str(result).endswith("test.txt")
    assert result.exists()
    t0 = result.stat().st_mtime_ns
    await asyncio.sleep(4)  # hopefully long enough for Windows/FAT32 ... ?

    result2 = await minio_adapter.download_file("test.txt", local_dir="/tmp/foo")
    assert result2.exists()
    assert result2 == result
    assert t0 == result2.stat().st_mtime_ns
    result.unlink()  # it should exist, from above ...


@pytest.mark.parametrize("populate_bucket", ["test.txt"], indirect=True)
@pytest.mark.asyncio
async def test_download_file_retry(download_patch, minio_adapter, populate_bucket):
    result = await minio_adapter.download_file("test.txt", local_dir="/tmp/foo")
    assert str(result).endswith("test.txt")
    assert result.exists()
    assert len(download_patch.call_args_list) == 3
    result.unlink()


@pytest.mark.asyncio
async def test_get_signed_url(minio_adapter, moto_services):
    result = await minio_adapter.get_signed_url("test.txt")
    assert result.startswith(moto_services["s3"])



# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from unittest.mock import MagicMock, patch

import pytest
from pytest_asyncio import fixture

from servicex.models import TransformedResults, ResultFormat
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter
from servicex.servicex_client import ServiceXClient


@fixture
def servicex_adaptor(mocker):
    adapter_mock = mocker.patch("servicex.servicex_client.ServiceXAdapter")
    mock_adapter = MagicMock(spec=ServiceXAdapter)

    adapter_mock.return_value = mock_adapter
    return mock_adapter


@fixture
def mock_cache(mocker):
    cache_mock = mocker.patch("servicex.servicex_client.QueryCache")
    mock_cache = MagicMock(spec=QueryCache)
    mock_cache.get_codegen_by_backend.return_value = {
        "codegens": {"ROOT": "my_root_generator", "UPROOT": "my_uproot_generator"}
    }
    cache_mock.return_value = mock_cache
    return cache_mock


def test_get_datasets(mock_cache, servicex_adaptor):
    sx = ServiceXClient(config_path="tests/example_config.yaml")
    sx.get_datasets()
    servicex_adaptor.get_datasets.assert_called_once()


def test_get_dataset(mock_cache, servicex_adaptor):
    sx = ServiceXClient(config_path="tests/example_config.yaml")
    sx.get_dataset("123")
    servicex_adaptor.get_dataset.assert_called_once_with("123")


def test_delete_dataset(mock_cache, servicex_adaptor):
    sx = ServiceXClient(config_path="tests/example_config.yaml")
    sx.delete_dataset("123")
    servicex_adaptor.delete_dataset.assert_called_once_with("123")


def test_delete_transform(mock_cache, servicex_adaptor):
    sx = ServiceXClient(config_path="tests/example_config.yaml")
    sx.delete_transform("123-45-6789")
    servicex_adaptor.delete_transform.assert_called_once_with("123-45-6789")


def test_cancel_transform(mock_cache, servicex_adaptor):
    sx = ServiceXClient(config_path="tests/example_config.yaml")
    sx.cancel_transform("123-45-6789")
    servicex_adaptor.cancel_transform.assert_called_once_with("123-45-6789")


@pytest.fixture
def transformed_results() -> TransformedResults:
    """
    Pytest fixture generating a realistic TransformedResults instance for testing.

    Returns:
        TransformedResults: A populated TransformedResults object with sample data
    """
    # Default data that can be overridden
    base_data = {
        "hash": "abc123def456",
        "title": "Test Transformation Request",
        "codegen": "python-selector",
        "request_id": "servicex-request-789",
        "submit_time": datetime.now(),
        "data_dir": "/tmp/servicex/results",
        "file_list": [
            "/tmp/servicex/results/output1.parquet",
            "/tmp/servicex/results/output2.parquet",
        ],
        "signed_url_list": [
            "https://example.com/signed-url-1",
            "https://example.com/signed-url-2",
        ],
        "files": 2,
        "result_format": ResultFormat.parquet,
        "log_url": "https://logs.servicex.com/request-789",
    }

    return TransformedResults(**base_data)


def test_delete_transform_from_cache(mock_cache, servicex_adaptor, transformed_results):
    with patch("servicex.servicex_client.QueryCache") as mock_cache:
        mock_cache.return_value.get_transform_by_request_id = MagicMock(
            return_value=transformed_results
        )
        with patch("servicex.servicex_client.shutil.rmtree") as mock_rmtree:
            sx = ServiceXClient(config_path="tests/example_config.yaml")
            sx.delete_transform_from_cache("servicex-request-789")

            mock_cache.return_value.get_transform_by_request_id.assert_called_once_with(
                "servicex-request-789"
            )
            mock_rmtree.assert_called_once_with(
                "/tmp/servicex/results", ignore_errors=True
            )
            mock_cache.return_value.delete_record_by_request_id.assert_called_once_with(
                "servicex-request-789"
            )



from servicex.app.transforms import LogLevel, TimeFrame
from servicex.app.transforms import (
    add_query,
    select_time,
    create_kibana_link_parameters,
)


def test_add_query():
    key = "abc"
    value = "123-345-567"
    query = "(query:(match_phrase:(abc:'123-345-567')))"
    assert add_query(key, value) == query

    key = "requestId"
    value = "d2ede739-9779-4075-95b1-0c7fae1de408"
    query = "(query:(match_phrase:(requestId:'d2ede739-9779-4075-95b1-0c7fae1de408')))"
    assert add_query(key, value) == query


def test_select_time():
    time_frame = TimeFrame.week
    time_filter = "time:(from:now%2Fw,to:now%2Fw)"
    assert time_filter == select_time(time_frame)

    time_frame = "month"
    time_filter = "time:(from:now-30d%2Fd,to:now)"
    assert time_filter == select_time(time_frame)

    time_frame = "daY"
    time_filter = "time:(from:now%2Fd,to:now%2Fd)"
    assert time_filter == select_time(time_frame)


def test_create_kibana_link_parameters():
    initial_log_url = (
        "https://atlas-kibana.mwt2.org:5601/s/servicex/app"
        "/dashboards?auth_provider_hint=anonymous1#/view/"
        "2d2b3b40-f34e-11ed-a6d8-9f6a16cd6d78?embed=true&_g=()"
        "&show-time-filter=true&hide-filter-bar=true"
    )
    transform_id = "d2ede739-9779-4075-95b1-0c7fae1de408"
    log_level = LogLevel.error
    time_frame = TimeFrame.day
    final_url = (
        "https://atlas-kibana.mwt2.org:5601/s/servicex/app/dashboards?"
        "auth_provider_hint=anonymous1#/view/2d2b3b40-f34e-11ed-a6d8-9f6a16cd6d78?"
        "embed=true&_g=(time:(from:now%2Fd,to:now%2Fd))"
        "&_a=(filters:!((query:(match_phrase:"
        "(requestId:'d2ede739-9779-4075-95b1-0c7fae1de408'))),"
        "(query:(match_phrase:(level:'error')))))&show-time-filter=true"
        "&hide-filter-bar=true"
    )
    assert (
        create_kibana_link_parameters(
            initial_log_url, transform_id, log_level, time_frame
        )
        == final_url
    )

    transform_id = "93713b34-2f0b-4d53-8412-8afa98626516"
    log_level = LogLevel.info
    time_frame = TimeFrame.month
    final_url = (
        "https://atlas-kibana.mwt2.org:5601/s/servicex/app/dashboards?"
        "auth_provider_hint=anonymous1#/view/2d2b3b40-f34e-11ed-a6d8-9f6a16cd6d78?"
        "embed=true&_g=(time:(from:now-30d%2Fd,to:now))"
        "&_a=(filters:!((query:(match_phrase:"
        "(requestId:'93713b34-2f0b-4d53-8412-8afa98626516'))),"
        "(query:(match_phrase:(level:'info')))))&show-time-filter=true"
        "&hide-filter-bar=true"
    )
    assert (
        create_kibana_link_parameters(
            initial_log_url, transform_id, log_level, time_frame
        )
        == final_url
    )

    transform_id = "93713b34-2f0b-4d53-8412-8afa98626516"
    log_level = None
    time_frame = TimeFrame.month
    final_url = (
        "https://atlas-kibana.mwt2.org:5601/s/servicex/app/dashboards?"
        "auth_provider_hint=anonymous1#/view/2d2b3b40-f34e-11ed-a6d8-9f6a16cd6d78?"
        "embed=true&_g=(time:(from:now-30d%2Fd,to:now))"
        "&_a=(filters:!((query:(match_phrase:"
        "(requestId:'93713b34-2f0b-4d53-8412-8afa98626516')))))"
        "&show-time-filter=true&hide-filter-bar=true"
    )
    assert (
        create_kibana_link_parameters(
            initial_log_url, transform_id, log_level, time_frame
        )
        == final_url
    )



import pytest
from unittest.mock import patch
from pydantic import ValidationError

from servicex import ServiceXSpec, dataset, OutputFormat
from servicex.query_core import ServiceXException
from servicex.servicex_client import ReturnValueException
from servicex.dataset import FileList, Rucio


def basic_spec(samples=None):
    return {
        "Sample": samples
        or [{"Name": "sampleA", "XRootDFiles": "root://a.root", "Query": "a"}],
    }


def test_long_sample_name():
    config = {
        "Sample": [
            {
                "Name": "long_sample_name_long_sample_name_long_sample_name_\
                         long_sample_name_long_sample_name_long_sample_name_\
                         long_sample_name_long_sample_name_long_sample_name",
                "XRootDFiles": "root://a.root",
                "Query": "a",
            },
        ],
    }
    new_config = ServiceXSpec.model_validate(config)
    assert len(new_config.Sample[0].Name) == 128


def test_load_config():
    config = {
        "General": {
            "Codegen": "python",
            "Delivery": "LocalCache",
        },
        "Sample": [
            {"Name": "sampleA", "RucioDID": "user.kchoi:sampleA", "Query": "a"},
            {
                "Name": "sampleB",
                "XRootDFiles": "root://a.root",
                "Query": "a",
            },
        ],
    }
    new_config = ServiceXSpec.model_validate(config)
    assert new_config.Sample[0].Query == "a"


def test_single_root_file():
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "XRootDFiles": "root://eospublic.cern.ch//file1.root",
                    "Query": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, FileList)
    assert spec.Sample[0].dataset_identifier.files == [
        "root://eospublic.cern.ch//file1.root"
    ]
    assert spec.Sample[0].dataset_identifier.did is None


def test_list_of_root_files():
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "XRootDFiles": [
                        "root://eospublic.cern.ch//file1.root",
                        "root://eospublic.cern.ch//file2.root",
                    ],
                    "Query": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, FileList)
    assert spec.Sample[0].dataset_identifier.files == [
        "root://eospublic.cern.ch//file1.root",
        "root://eospublic.cern.ch//file2.root",
    ]


def test_output_format():
    spec = basic_spec()
    spec["General"] = {"OutputFormat": "root-ttree"}
    ServiceXSpec.model_validate(spec)
    spec["General"] = {"OutputFormat": "parquet"}
    ServiceXSpec.model_validate(spec)
    spec["General"] = {"OutputFormat": OutputFormat.root_ttree}
    ServiceXSpec.model_validate(spec)
    spec["General"] = {"OutputFormat": OutputFormat.parquet}
    ServiceXSpec.model_validate(spec)
    with pytest.raises(ValidationError):
        spec["General"] = {"OutputFormat": "root-tree"}
        ServiceXSpec.model_validate(spec)


def test_rucio_did():
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, Rucio)
    assert (
        spec.Sample[0].dataset_identifier.did
        == "rucio://user.ivukotic:user.ivukotic.single_top_tW__nominal"
    )


def test_rucio_did_numfiles():
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "NFiles": 10,
                    "Query": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, Rucio)
    assert (
        spec.Sample[0].dataset_identifier.did
        == "rucio://user.ivukotic:user.ivukotic.single_top_tW__nominal?files=10"
    )


def test_dataset_rucio_did_numfiles():
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "Dataset": dataset.Rucio(
                        "user.ivukotic:user.ivukotic.single_top_tW__nominal"
                    ),
                    "NFiles": 12,
                    "Query": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, Rucio)
    assert (
        spec.Sample[0].dataset_identifier.did
        == "rucio://user.ivukotic:user.ivukotic.single_top_tW__nominal?files=12"
    )


def test_dataset_zerofiles():
    # with an actual dataset, giving no files should throw a validation error
    with pytest.raises(ValidationError):
        spec = ServiceXSpec.model_validate(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "Dataset": dataset.Rucio(
                            "user.ivukotic:user.ivukotic.single_top_tW__nominal"
                        ),
                        "NFiles": 0,
                        "Query": "a",
                    }
                ]
            )
        )

    with pytest.raises(ValidationError):
        spec = ServiceXSpec.model_validate(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "Dataset": dataset.Rucio(
                            "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                            num_files=0,
                        ),
                        "Query": "a",
                    }
                ]
            )
        )

    # and num files should be ignored for fileset
    spec = ServiceXSpec.model_validate(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "Dataset": dataset.FileList(
                        [
                            "root://eospublic.cern.ch//file1.root",
                            "root://eospublic.cern.ch//file2.root",
                        ]
                    ),
                    "Query": "a",
                }
            ]
        )
    )
    assert spec.Sample[0].dataset_identifier.num_files is None


def test_cernopendata():
    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "Dataset": dataset.CERNOpenData(1507),
                    "Function": "a",
                }
            ]
        }
    )
    assert spec.Sample[0].dataset_identifier.did == "cernopendata://1507"


def test_xrootd():
    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "Dataset": dataset.XRootD("root://blablabla/*/?.root"),
                    "Function": "a",
                }
            ]
        }
    )
    assert spec.Sample[0].dataset_identifier.did == "xrootd://root://blablabla/*/?.root"


def test_invalid_dataset_identifier():
    with pytest.raises(ValidationError):
        ServiceXSpec.model_validate(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                        "XRootDFiles": "root://eospublic.cern.ch//file1.root",
                        "NFiles": 10,
                        "Query": "a",
                    }
                ]
            )
        )

    with pytest.raises(ValidationError):
        ServiceXSpec.model_validate(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "NFiles": 10,
                        "Query": "a",
                    }
                ]
            )
        )


def test_submit_mapping(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver

    spec = {
        "General": {
            "Codegen": "uproot-raw",
        },
        "Sample": [
            {
                "Name": "sampleA",
                "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                "Query": "[{'treename': 'nominal'}]",
                "Codegen": "uproot-raw",
            }
        ],
    }
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        results = deliver(spec, config_path="tests/example_config.yaml")
        assert list(results["sampleA"]) == ["1.parquet"]


def test_submit_mapping_signed_urls(
    transformed_result_signed_url, codegen_list, with_event_loop
):
    from servicex import deliver

    spec = {
        "General": {"Delivery": "URLs"},
        "Sample": [
            {
                "Name": "sampleA",
                "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                "Query": "[{'treename': 'nominal'}]",
                "Codegen": "uproot-raw",
            }
        ],
    }
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_signed_urls",
            return_value=[transformed_result_signed_url],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        results = deliver(spec, config_path="tests/example_config.yaml")
        assert list(results["sampleA"]) == [
            "https://dummy.junk.io/1.parquet",
            "https://dummy.junk.io/2.parquet",
        ]


def test_submit_mapping_failure(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver

    spec = {
        "Sample": [
            {
                "Name": "sampleA",
                "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                "Query": "[{'treename': 'nominal'}]",
                "Codegen": "uproot-raw",
            }
        ]
    }
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[ServiceXException("dummy")],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        results = deliver(spec, config_path="tests/example_config.yaml")
        assert len(results) == 1
        with pytest.raises(ReturnValueException):
            # should expect an exception to be thrown on access
            for _ in results["sampleA"]:
                pass


def test_submit_mapping_failure_signed_urls(codegen_list, with_event_loop):
    from servicex import deliver

    spec = {
        "General": {"Delivery": "URLs"},
        "Sample": [
            {
                "Name": "sampleA",
                "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                "Query": "[{'treename': 'nominal'}]",
                "Codegen": "uproot-raw",
            }
        ],
    }
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_signed_urls",
            return_value=[ServiceXException("dummy")],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        results = deliver(
            spec, config_path="tests/example_config.yaml", return_exceptions=False
        )
        assert len(results) == 1
        with pytest.raises(ReturnValueException):
            # should expect an exception to be thrown on access
            for _ in results["sampleA"]:
                pass


def test_yaml(tmp_path):
    from servicex.servicex_client import _load_ServiceXSpec
    from servicex.dataset import FileList, Rucio, CERNOpenData

    # Nominal paths
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !PythonFunction |
        def run_query(input_filenames=None):
            return []
    Codegen: python
  - Name: ttH2
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !FuncADL_Uproot |
                Select(lambda e: {'lep_pt': e['lep_pt']})
                .FromTree("nominal")
  - Name: ttH3
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !UprootRaw |
                [{"treename": "nominal"}]
  - Name: ttH4
    Dataset: !Rucio user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v112
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH5
    Dataset: !FileList ["/path/to/file1.root", "/path/to/file2.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH6
    Dataset: !CERNOpenData 1507
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH7
    Dataset: !XRootD root://eosatlas.cern.ch//eos/atlas/path/*/file.root
"""
        )
        f.flush()
        result = _load_ServiceXSpec(path)
        assert type(result.Sample[0].Query).__name__ == "PythonFunction"
        assert type(result.Sample[1].Query).__name__ == "FuncADLQuery_Uproot"
        assert type(result.Sample[2].Query).__name__ == "UprootRawQuery"
        assert isinstance(result.Sample[3].dataset_identifier, Rucio)
        assert (
            result.Sample[3].dataset_identifier.did
            == "rucio://user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v112"
        )
        assert isinstance(result.Sample[4].dataset_identifier, FileList)
        assert result.Sample[4].dataset_identifier.files == [
            "/path/to/file1.root",
            "/path/to/file2.root",
        ]
        assert isinstance(result.Sample[5].dataset_identifier, CERNOpenData)
        assert result.Sample[5].dataset_identifier.did == "cernopendata://1507"
        assert (
            result.Sample[6].dataset_identifier.did
            == "xrootd://root://eosatlas.cern.ch//eos/atlas/path/*/file.root"
        )

    # Path from string
    result2 = _load_ServiceXSpec(str(path))
    assert type(result2.Sample[0].Query).__name__ == "PythonFunction"

    # Python syntax error
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !PythonFunction |
        def run_query(input_filenames=None):
            i ==== 18 # syntax error
"""
        )
        f.flush()
        with pytest.raises(SyntaxError):
            _load_ServiceXSpec(path)

    # Duplicate samples with different names but same dataset and query
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH3
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !UprootRaw |
                [{"treename": "nominal"}]
  - Name: ttH4
    Dataset: !Rucio user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH5
    Dataset: !FileList ["/path/to/file1.root", "/path/to/file2.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
    """
        )
        f.flush()
        with pytest.raises(RuntimeError):
            _load_ServiceXSpec(path)

    # Duplicate samples with different names but same datasets (multiple) and query
    # change the order of the datasets
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH3
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: !UprootRaw |
                [{"treename": "nominal"}]
  - Name: ttH5
    Dataset: !FileList ["/path/to/file2.root", "/path/to/file1.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH5
    Dataset: !FileList ["/path/to/file1.root", "/path/to/file2.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
    """
        )
        f.flush()
        with pytest.raises(RuntimeError):
            _load_ServiceXSpec(path)

    # Samples with different names but same datasets(multiple) and query
    # different NFiles
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH5
    Dataset: !FileList ["/path/to/file2.root", "/path/to/file1.root"]
    NFiles: 3
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH6
    NFiles: 1
    Dataset: !FileList ["/path/to/file1.root", "/path/to/file2.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
    """
        )
        f.flush()
        result = _load_ServiceXSpec(path)
        assert type(result.Sample[0].Query).__name__ == "UprootRawQuery"
        assert type(result.Sample[1].Query).__name__ == "UprootRawQuery"

    # Samples with different names but same datasets(multiple) and
    # different queries
    with open(path := (tmp_path / "python.yaml"), "w") as f:
        f.write(
            """
General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH5
    Dataset: !FileList ["/path/to/file2.root", "/path/to/file1.root"]
    Query: !UprootRaw '[{"treename": "nominal"}]'
  - Name: ttH6
    Dataset: !FileList ["/path/to/file1.root", "/path/to/file2.root"]
    Query: !UprootRaw '[{"treename": "CollectionTree"}]'
    """
        )
        f.flush()
        result = _load_ServiceXSpec(path)
        assert type(result.Sample[0].Query).__name__ == "UprootRawQuery"
        assert type(result.Sample[1].Query).__name__ == "UprootRawQuery"


def test_yaml_include(tmp_path):
    from servicex.servicex_client import _load_ServiceXSpec

    # Create two files, one has definitions for the other and is included by it
    with (
        open(tmp_path / "definitions.yaml", "w") as f1,
        open(path2 := (tmp_path / "parent.yaml"), "w") as f2,
    ):
        f1.write(
            """
- &DEF_query !PythonFunction |
        def run_query(input_filenames=None):
            return []
"""
        )
        f2.write(
            """
Definitions:
    !include definitions.yaml

General:
  OutputFormat: root-ttree
  Delivery: LocalCache

Sample:
  - Name: ttH
    RucioDID: user.kchoi:user.kchoi.fcnc_tHq_ML.ttH.v11
    Query: *DEF_query
"""
        )
        f1.flush()
        f2.flush()
        _load_ServiceXSpec(path2)


def test_funcadl_query(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver
    from servicex.query import FuncADL_Uproot  # type: ignore

    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": FuncADL_Uproot()
                    .FromTree("nominal")
                    .Select(lambda e: {"lep_pt": e["lep_pt"]}),
                }
            ]
        }
    )
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        deliver(spec, config_path="tests/example_config.yaml")


def test_query_with_codegen_override(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver
    from servicex.query import FuncADL_Uproot  # type: ignore

    # first, with General override
    spec = ServiceXSpec.model_validate(
        {
            "General": {"Codegen": "does-not-exist"},
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": FuncADL_Uproot()
                    .FromTree("nominal")
                    .Select(lambda e: {"lep_pt": e["lep_pt"]}),
                }
            ],
        }
    )
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        with pytest.raises(NameError) as excinfo:
            deliver(spec, config_path="tests/example_config.yaml")
        # if this has propagated correctly, the override worked
        assert excinfo.value.args[0].startswith("does-not-exist")

    # second, with sample-level override
    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": FuncADL_Uproot()
                    .FromTree("nominal")
                    .Select(lambda e: {"lep_pt": e["lep_pt"]}),
                    "Codegen": "does-not-exist",
                }
            ]
        }
    )
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        with pytest.raises(NameError) as excinfo:
            deliver(spec, config_path="tests/example_config.yaml")
        # if this has propagated correctly, the override worked
        assert excinfo.value.args[0].startswith("does-not-exist")


def test_databinder_load_dict():
    from servicex.query import FuncADL_Uproot  # type: ignore
    from servicex.servicex_client import _load_ServiceXSpec

    _load_ServiceXSpec(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": FuncADL_Uproot()
                    .FromTree("nominal")
                    .Select(lambda e: {"lep_pt": e["lep_pt"]}),
                }
            ]
        }
    )


def test_python_query(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver
    from servicex.query import PythonFunction  # type: ignore

    def run_query(input_filenames=None):
        print("Greetings from your query")
        return []

    query = PythonFunction().with_uproot_function(run_query)

    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": query,
                }
            ]
        }
    )
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        deliver(spec, config_path="tests/example_config.yaml")


def test_uproot_raw_query(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver
    from servicex.query import UprootRaw  # type: ignore

    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": UprootRaw([{"treename": "nominal"}]),
                }
            ]
        }
    )
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        deliver(spec, config_path="tests/example_config.yaml")


def test_uproot_raw_query_parquet(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver
    from servicex.query import UprootRaw  # type: ignore

    spec = ServiceXSpec.model_validate(
        {
            "General": {"OutputFormat": "parquet"},
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": UprootRaw([{"treename": "nominal"}]),
                }
            ],
        }
    )
    print(spec)
    with (
        patch(
            "servicex.dataset_group.DatasetGroup.as_files",
            return_value=[transformed_result],
        ),
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
    ):
        deliver(spec, config_path="tests/example_config.yaml")


def test_generic_query(codegen_list):
    from servicex.servicex_client import ServiceXClient

    spec = ServiceXSpec.model_validate(
        {
            "General": {
                "Codegen": "uproot-raw",
            },
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": "[{'treename': 'nominal'}]",
                }
            ],
        }
    )
    with patch(
        "servicex.servicex_client.ServiceXClient.get_code_generators",
        return_value=codegen_list,
    ):
        sx = ServiceXClient(config_path="tests/example_config.yaml")
        query = sx.generic_query(
            dataset_identifier=spec.Sample[0].RucioDID,
            codegen=spec.General.Codegen,
            query=spec.Sample[0].Query,
        )
        assert query.generate_selection_string() == "[{'treename': 'nominal'}]"
        query = sx.generic_query(
            dataset_identifier=spec.Sample[0].RucioDID,
            result_format=spec.General.OutputFormat.to_ResultFormat(),
            codegen=spec.General.Codegen,
            query=spec.Sample[0].Query,
        )
        assert query.result_format == "root-file"
        query.query_string_generator = None
        with pytest.raises(RuntimeError):
            query.generate_selection_string()
        with pytest.raises(ValueError):
            query = sx.generic_query(
                dataset_identifier=spec.Sample[0].RucioDID,
                codegen=spec.General.Codegen,
                query=5,
            )
        with pytest.raises(NameError):
            query = sx.generic_query(
                dataset_identifier=spec.Sample[0].RucioDID,
                codegen="nonsense",
                query=spec.Sample[0].Query,
            )
        with pytest.raises(RuntimeError):
            # no codegen specified by generic class
            query = sx.generic_query(
                dataset_identifier=spec.Sample[0].RucioDID, query=spec.Sample[0].Query
            )


def test_deliver_progress_options(transformed_result, codegen_list, with_event_loop):
    from servicex import deliver, ProgressBarFormat
    from servicex.query import UprootRaw  # type: ignore

    spec = ServiceXSpec.model_validate(
        {
            "Sample": [
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Query": UprootRaw([{"treename": "nominal"}]),
                }
            ]
        }
    )

    async def fake_submit(signed_urls_only, expandable_progress):
        expandable_progress.add_task("zip", start=False, total=None)
        return transformed_result

    with (
        patch(
            "servicex.servicex_client.ServiceXClient.get_code_generators",
            return_value=codegen_list,
        ),
        patch(
            "servicex.servicex_adapter.ServiceXAdapter._get_authorization",
            return_value={"Authorization": "Bearer aaa"},
        ),
        patch(
            "servicex.query_core.Query.submit_and_download",
            side_effect=fake_submit,
        ),
    ):
        import servicex.query_core

        rv = deliver(
            spec,
            config_path="tests/example_config.yaml",
            progress_bar=ProgressBarFormat.compact,
        )
        servicex.query_core.Query.submit_and_download.assert_called_once()
        assert rv is not None
        assert rv["sampleA"].valid()
        rv = deliver(
            spec,
            config_path="tests/example_config.yaml",
            progress_bar=ProgressBarFormat.none,
        )
        assert rv is not None
        assert rv["sampleA"].valid()
        deliver(
            spec,
            config_path="tests/example_config.yaml",
            progress_bar=ProgressBarFormat.expanded,
        )
        assert rv is not None
        assert rv["sampleA"].valid()
        with pytest.raises(ValueError):
            deliver(
                spec,
                config_path="tests/example_config.yaml",
                progress_bar="garbage",
            )


def test_entrypoint_import():
    """This will check that we have at least the Python transformer defined in servicex.query"""
    from servicex.query import PythonFunction  # type: ignore # noqa: F401



# Copyright (c) 2022-2025, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
import shutil
from typing import Optional, List, TypeVar, Any, Mapping, Union, cast
from pathlib import Path

from servicex.configuration import Configuration
from servicex.models import ResultFormat, TransformStatus, TransformedResults
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter
from servicex.query_core import (
    Query,
    QueryStringGenerator,
    GenericQueryStringGenerator,
)
from servicex.types import DID
from servicex.dataset_group import DatasetGroup

from make_it_sync import make_sync
from servicex.databinder_models import ServiceXSpec, General, Sample
from collections.abc import Sequence
from enum import Enum
import traceback

T = TypeVar("T")
logger = logging.getLogger(__name__)


class ProgressBarFormat(str, Enum):
    """Specify the way progress bars are displayed."""

    expanded = "expanded"
    """Show progress bars for each `Sample`"""
    compact = "compact"
    """Show one overall summary set of progress bars"""
    none = "none"
    """Show no progress bars at all"""
    default = "expanded"
    """Default (currently the same as "expanded")"""


class ReturnValueException(Exception):
    """An exception occurred at some point while obtaining this result from ServiceX"""

    def __init__(self, exc):
        import copy

        message = "Exception occurred while making ServiceX request.\n" + (
            "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
        )
        super().__init__(message)
        self._exc = copy.copy(exc)


class GuardList(Sequence):
    def __init__(self, data: Union[Sequence, Exception]):
        import copy

        super().__init__()
        if isinstance(data, Exception):
            self._data = ReturnValueException(data)
        else:
            self._data = copy.copy(data)

    def valid(self) -> bool:
        return not isinstance(self._data, Exception)

    def __getitem__(self, index) -> Any:
        if not self.valid():
            data = cast(Exception, self._data)
            raise data
        else:
            data = cast(Sequence, self._data)
            return data[index]

    def __len__(self) -> int:
        if not self.valid():
            data = cast(Exception, self._data)
            raise data
        else:
            data = cast(Sequence, self._data)
            return len(data)

    def __repr__(self):
        if self.valid():
            return repr(self._data)
        else:
            data = cast(ReturnValueException, self._data)
            return f"Invalid GuardList: {repr(data._exc)}"


def _load_ServiceXSpec(
    config: Union[ServiceXSpec, Mapping[str, Any], str, Path],
) -> ServiceXSpec:
    if isinstance(config, Mapping):
        logger.debug("Config from dictionary")
        config = ServiceXSpec(**config)
    elif isinstance(config, ServiceXSpec):
        logger.debug("Config from ServiceXSpec")
    elif isinstance(config, str) or isinstance(config, Path):
        logger.debug("Config from file")

        if isinstance(config, str):
            file_path = Path(config)
        else:
            file_path = config

        import sys
        from .yaml_parser import YAML

        yaml = YAML()

        if sys.version_info < (3, 10):
            from importlib_metadata import entry_points
        else:
            from importlib.metadata import entry_points

        plugins = entry_points(group="servicex.query")
        for _ in plugins:
            yaml.register_class(_.load())
        plugins = entry_points(group="servicex.dataset")
        for _ in plugins:
            yaml.register_class(_.load())

        conf = yaml.load(file_path)
        config = ServiceXSpec(**conf)
    else:
        raise TypeError(f"Unknown config type: {type(config)}")

    return config


def _build_datasets(config, config_path, servicex_name, fail_if_incomplete):
    def get_codegen(_sample: Sample, _general: General):
        if _sample.Codegen is not None:
            return _sample.Codegen
        elif _general.Codegen is not None:
            return _general.Codegen
        elif isinstance(_sample.Query, QueryStringGenerator):
            return _sample.Query.default_codegen
        elif isinstance(_sample.Query, Query):
            return _sample.Query.codegen

    sx = ServiceXClient(backend=servicex_name, config_path=config_path)
    datasets = []
    for sample in config.Sample:
        query = sx.generic_query(
            dataset_identifier=sample.dataset_identifier,
            title=sample.Name,
            codegen=get_codegen(sample, config.General),
            result_format=config.General.OutputFormat.to_ResultFormat(),
            ignore_cache=sample.IgnoreLocalCache,
            query=sample.Query,
            fail_if_incomplete=fail_if_incomplete,
        )
        logger.debug(f"Query string: {query.generate_selection_string()}")
        query.ignore_cache = sample.IgnoreLocalCache

        datasets.append(query)
    return datasets


def _output_handler(
    config: ServiceXSpec,
    requests: List[Query],
    results: List[Union[TransformedResults, Exception]],
):
    matched_results = zip(requests, results)
    if config.General.Delivery == General.DeliveryEnum.URLs:
        out_dict = {
            obj[0].title: GuardList(
                obj[1].signed_url_list if not isinstance(obj[1], Exception) else obj[1]
            )
            for obj in matched_results
        }
    elif config.General.Delivery == General.DeliveryEnum.LocalCache:
        out_dict = {
            obj[0].title: GuardList(
                obj[1].file_list if not isinstance(obj[1], Exception) else obj[1]
            )
            for obj in matched_results
        }

    if config.General.OutputDirectory:
        import yaml as yl

        out_dir = Path(config.General.OutputDirectory).absolute()
        out_dir.mkdir(parents=True, exist_ok=True)
        out_dict_path = Path(out_dir, config.General.OutFilesetName + ".yaml")

        with open(out_dict_path, "w") as f:
            yl.dump(out_dict, f, default_flow_style=False)

    return out_dict


def deliver(
    spec: Union[ServiceXSpec, Mapping[str, Any], str, Path],
    config_path: Optional[str] = None,
    servicex_name: Optional[str] = None,
    return_exceptions: bool = True,
    fail_if_incomplete: bool = True,
    ignore_local_cache: bool = False,
    progress_bar: ProgressBarFormat = ProgressBarFormat.default,
    concurrency: int = 10,
):
    r"""
    Execute a ServiceX query.

    :param spec: The specification of the ServiceX query, either in a dictionary or a
            :py:class:`~servicex.ServiceXSpec` object.
    :param config_path: The filesystem path to search for the `servicex.yaml` or `.servicex` file.
    :param servicex_name: The name of the ServiceX instance, as specified in the configuration
            YAML file (None will give the default backend).
    :param return_exceptions: If something goes wrong, bubble up the underlying exception for
            debugging (as opposed to just having a generic error).
    :param fail_if_incomplete: If :py:const:`True`: if not all input files are transformed, the
            transformation will be marked as a failure and no outputs will be available. If
            :py:const:`False`, a partial file list will be returned.
    :param ignore_local_cache: If :py:const:`True`, ignore the local query cache and always run
            the query on the remote ServiceX instance.
    :param progress_bar: specify the kind of progress bar to show.
            :py:const:`ProgressBarFormat.expanded` (the default) means every :py:class:`Sample`
            will have its own progress bars; :py:const:`ProgressBarFormat.compact` gives one
            summary progress bar for all transformations; :py:const:`ProgressBarFormat.none`
            switches off progress bars completely.
    :param concurrency: specify how many downloads to run in parallel (default is 8).
    :return: A dictionary mapping the name of each :py:class:`Sample` to a :py:class:`.GuardList`
            with the file names or URLs for the outputs.
    """
    from .minio_adapter import init_s3_config

    init_s3_config(concurrency)
    config = _load_ServiceXSpec(spec)

    if ignore_local_cache or config.General.IgnoreLocalCache:
        for sample in config.Sample:
            sample.IgnoreLocalCache = True

    datasets = _build_datasets(config, config_path, servicex_name, fail_if_incomplete)

    group = DatasetGroup(datasets)

    if progress_bar == ProgressBarFormat.expanded:
        progress_options = {}
    elif progress_bar == ProgressBarFormat.compact:
        progress_options = {"overall_progress": True}
    elif progress_bar == ProgressBarFormat.none:
        progress_options = {"display_progress": False}
    else:
        raise ValueError(f"Invalid value {progress_bar} for progress_bar provided")

    if config.General.Delivery == General.DeliveryEnum.URLs:
        results = group.as_signed_urls(
            return_exceptions=return_exceptions, **progress_options
        )
        return _output_handler(config, datasets, results)

    elif config.General.Delivery == General.DeliveryEnum.LocalCache:
        results = group.as_files(
            return_exceptions=return_exceptions, **progress_options
        )
        return _output_handler(config, datasets, results)


class ServiceXClient:
    r"""
    Connection to a ServiceX deployment. Instances of this class can deployment
    data from the service and also interact with previously run transformations.
    Instances of this class are factories for `Datasets``
    """

    def __init__(self, backend=None, url=None, config_path=None):
        r"""
        If both `backend` and `url` are unspecified then it will attempt to pick up
        the default backend from `.servicex`

        :param backend: Name of a deployment from the .servicex file
        :param url: Direct URL of a serviceX deployment instead of using .servicex.
                    Can only work with hosts without auth, or the token is found
                    in a file pointed to by the environment variable BEARER_TOKEN_FILE
        :param config_path: Optional path to the `.servicex` file. If not specified,
                    will search in local directory and up in enclosing directories
        """
        self.config = Configuration.read(config_path)
        self.endpoints = self.config.endpoint_dict()

        if not url and not backend:
            if self.config.default_endpoint:
                backend = self.config.default_endpoint
            else:
                # Take the first endpoint from servicex.yaml if default_endpoint is not set
                backend = self.config.api_endpoints[0].name

        if bool(url) == bool(backend):
            raise ValueError("Only specify backend or url... not both")

        if url:
            self.servicex = ServiceXAdapter(url)
        elif backend:
            if backend not in self.endpoints:
                raise ValueError(f"Backend {backend} not defined in .servicex file")
            self.servicex = ServiceXAdapter(
                self.endpoints[backend].endpoint,
                refresh_token=self.endpoints[backend].token,
            )

        self.query_cache = QueryCache(self.config)
        self.code_generators = set(self.get_code_generators(backend).keys())

    async def get_transforms_async(self) -> List[TransformStatus]:
        r"""
        Retrieve all transforms you have run on the server
        :return: List of Transform status objects
        """
        return await self.servicex.get_transforms()

    get_transforms = make_sync(get_transforms_async)

    async def get_transform_status_async(self, transform_id) -> TransformStatus:
        r"""
        Get the status of a given transform
        :param transform_id: The uuid of the transform
        :return: The current status for the transform
        """
        return await self.servicex.get_transform_status(request_id=transform_id)

    get_transform_status = make_sync(get_transform_status_async)

    def get_datasets(self, did_finder=None, show_deleted=False):
        r"""
        Retrieve all datasets you have run on the server
        :return: List of Query objects
        """
        return self.servicex.get_datasets(did_finder, show_deleted)

    def get_dataset(self, dataset_id):
        r"""
        Retrieve a dataset by its ID
        :return: A Query object
        """
        return self.servicex.get_dataset(dataset_id)

    def delete_dataset(self, dataset_id):
        r"""
        Delete a dataset by its ID
        :return: A Query object
        """
        return self.servicex.delete_dataset(dataset_id)

    def delete_transform(self, transform_id):
        r"""
        Delete a Transform by its request ID
        :return: A Query object
        """
        return self.servicex.delete_transform(transform_id)

    def cancel_transform(self, transform_id):
        r"""
        Cancel a Transform by its request ID
        :return: A Query object
        """
        return self.servicex.cancel_transform(transform_id)

    def get_code_generators(self, backend=None):
        r"""
        Retrieve the code generators deployed with the serviceX instance
        :return:  The list of code generators as json dictionary
        """
        cached_backends = None
        if backend:
            cached_backends = self.query_cache.get_codegen_by_backend(backend)
        if cached_backends:
            logger.info("Returning code generators from cache")
            return cached_backends["codegens"]
        else:
            code_generators = self.servicex.get_code_generators()
            self.query_cache.update_codegen_by_backend(backend, code_generators)
            return code_generators

    def generic_query(
        self,
        dataset_identifier: DID,
        query: Union[str, QueryStringGenerator],
        codegen: Optional[str] = None,
        title: str = "ServiceX Client",
        result_format: ResultFormat = ResultFormat.parquet,
        ignore_cache: bool = False,
        fail_if_incomplete: bool = True,
    ) -> Query:
        r"""
        Generate a Query object for a generic codegen specification

        :param dataset_identifier:  The dataset identifier or filelist to be the source of files
        :param title: Title to be applied to the transform. This is also useful for
                      relating transform results.
        :param codegen: Name of the code generator to use with this transform
        :param result_format:  Do you want Paqrquet or Root? This can be set later with
                               the set_result_format method
        :param ignore_cache: Ignore the query cache and always run the query
        :return: A Query object

        """

        if isinstance(query, str):
            if codegen is None:
                raise RuntimeError(
                    "A pure string query requires a codegen argument as well"
                )
            query = GenericQueryStringGenerator(query, codegen)
        if not isinstance(query, QueryStringGenerator):
            raise ValueError("query argument must be string or QueryStringGenerator")

        real_codegen = codegen if codegen is not None else query.default_codegen
        if real_codegen is None:
            raise RuntimeError(
                "No codegen specified, either from query class or user input"
            )

        if real_codegen not in self.code_generators:
            raise NameError(
                f"{codegen} code generator not supported by serviceX "
                f"deployment at {self.servicex.url}"
            )

        qobj = Query(
            dataset_identifier=dataset_identifier,
            sx_adapter=self.servicex,
            title=title,
            codegen=real_codegen,
            config=self.config,
            query_cache=self.query_cache,
            result_format=result_format,
            ignore_cache=ignore_cache,
            query_string_generator=query,
            fail_if_incomplete=fail_if_incomplete,
        )
        return qobj

    def delete_transform_from_cache(self, transform_id: str):
        cache = self.query_cache
        rec = cache.get_transform_by_request_id(transform_id)
        if not rec:
            return False

        shutil.rmtree(rec.data_dir, ignore_errors=True)
        cache.delete_record_by_request_id(rec.request_id)
        return True



# Copyright (c) 2024-2025, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pydantic 2 API

import pydantic
from ..models import DocStringBaseModel
from typing import List, Union, Mapping, Optional, get_args
from ..query_core import QueryStringGenerator


class TreeSubQuery(DocStringBaseModel):
    """Express an uproot-raw query on an ntuple. Arguments have the same meaning as the
    ones of the same name for uproot.arrays():
    https://uproot.readthedocs.io/en/stable/uproot.behaviors.TBranch.HasBranches.html#uproot-behaviors-tbranch-hasbranches-arrays
    """

    model_config = {"use_attribute_docstrings": True}

    treename: Union[Mapping[str, str], List[str], str]
    """Name of input ntuple in file"""
    expressions: Optional[Union[List[str], str]] = None
    """Awkward Array expressions to evaluate for output"""
    cut: Optional[str] = None
    """Awkward Array expressions to select passing rows"""
    filter_name: Optional[Union[List[str], str]] = None
    """Expression to filter the list of considered input variables by name"""
    filter_typename: Optional[Union[List[str], str]] = None
    """Expression to filter the list of considered input variables by type name"""
    aliases: Optional[Mapping[str, str]] = None
    """Define aliases to use in computation and expressions"""
    fail_on_missing_trees: Optional[bool] = None
    """Make queries fail if input trees are missing (default False)"""


class CopyHistogramSubQuery(DocStringBaseModel):
    """Request the copying of a ROOT object from the input file to the output."""

    model_config = {"use_attribute_docstrings": True}

    copy_histograms: Union[List[str], str]
    """Objects to copy"""


SubQuery = Union[TreeSubQuery, CopyHistogramSubQuery]


@pydantic.dataclasses.dataclass
class UprootRawQuery(QueryStringGenerator):
    yaml_tag = "!UprootRaw"

    query: Union[List[SubQuery], SubQuery]
    default_codegen: str = "uproot-raw"

    def generate_selection_string(self):
        import json

        final_query: List[SubQuery]
        if isinstance(
            self.query, get_args(SubQuery)
        ):  # from Python 3.10 we don't need "get_args"
            final_query = [self.query]
        else:
            final_query = self.query
        return json.dumps([json.loads(_.model_dump_json()) for _ in final_query])

    @classmethod
    def from_yaml(cls, _, node):
        code = node.value
        import json

        queries = json.loads(code)
        q = cls(queries)
        return q



from servicex.servicex_client import GuardList, ReturnValueException
import pytest


def test_guardlist():
    gl1 = GuardList([1])
    assert str(gl1) == "[1]"
    assert gl1[0] == 1
    gl2 = GuardList(ValueError())
    assert str(gl2) == "Invalid GuardList: ValueError()"
    with pytest.raises(ReturnValueException):
        gl2[0]
    with pytest.raises(ReturnValueException):
        len(gl2)



from unittest.mock import patch

from servicex.servicex_client import ServiceXClient


def test_default_endpoint(codegen_list):
    with patch(
        "servicex.servicex_client.ServiceXClient.get_code_generators",
        return_value=codegen_list,
    ):
        sx = ServiceXClient(config_path="tests/example_config.yaml")
        assert sx.servicex.url == "http://localhost:5000"


def test_first_endpoint(codegen_list):
    with patch(
        "servicex.servicex_client.ServiceXClient.get_code_generators",
        return_value=codegen_list,
    ):
        sx = ServiceXClient(config_path="tests/example_config_default_endpoint.yaml")
        assert sx.servicex.url == "https://servicex.af.uchicago.edu"



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import time
from typing import Optional, Dict, List

from aiohttp import ClientSession
import httpx
from aiohttp_retry import RetryClient, ExponentialRetry, ClientResponse
from aiohttp import ContentTypeError
from google.auth import jwt
from tenacity import (
    AsyncRetrying,
    stop_after_attempt,
    wait_fixed,
    retry_if_not_exception_type,
)

from servicex.models import TransformRequest, TransformStatus, CachedDataset


class AuthorizationError(BaseException):
    pass


async def _extract_message(r: ClientResponse):
    try:
        o = await r.json()
        error_message = o.get("message", str(r))
    except ContentTypeError:
        error_message = await r.text()
    return error_message


class ServiceXAdapter:
    def __init__(self, url: str, refresh_token: Optional[str] = None):
        self.url = url
        self.refresh_token = refresh_token
        self.token = None

    async def _get_token(self):
        url = f"{self.url}/token/refresh"
        headers = {"Authorization": f"Bearer {self.refresh_token}"}
        async with RetryClient() as client:
            async with client.post(url, headers=headers, json=None) as r:
                if r.status == 200:
                    o = await r.json()
                    self.token = o["access_token"]
                else:
                    raise AuthorizationError(
                        f"ServiceX access token request rejected [{r.status} {r.reason}]"
                    )

    @staticmethod
    def _get_bearer_token_file():
        bearer_token_file = os.environ.get("BEARER_TOKEN_FILE")
        bearer_token = None
        if bearer_token_file:
            with open(bearer_token_file, "r") as f:
                bearer_token = f.read().strip()
        return bearer_token

    async def _get_authorization(self, force_reauth: bool = False) -> Dict[str, str]:
        now = time.time()
        if (
            self.token
            and jwt.decode(self.token, verify=False)["exp"] - now > 60
            and not force_reauth
        ):
            # if less than one minute validity, renew
            return {"Authorization": f"Bearer {self.token}"}
        else:
            bearer_token = self._get_bearer_token_file()

            if bearer_token:
                self.token = bearer_token
            if not bearer_token and not self.refresh_token:
                return {}

            if (
                not self.token
                or force_reauth
                or float(jwt.decode(self.token, verify=False)["exp"]) - now < 60
            ):
                await self._get_token()
            return {"Authorization": f"Bearer {self.token}"}

    async def get_transforms(self) -> List[TransformStatus]:
        headers = await self._get_authorization()
        retry_options = ExponentialRetry(attempts=3, start_timeout=10)
        async with RetryClient(retry_options=retry_options) as client:
            async with client.get(
                url=f"{self.url}/servicex/transformation", headers=headers
            ) as r:
                if r.status == 401:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status > 400:
                    error_message = await _extract_message(r)
                    raise RuntimeError(
                        "ServiceX WebAPI Error during transformation "
                        f"submission: {r.status} - {error_message}"
                    )
                o = await r.json()
                statuses = [TransformStatus(**status) for status in o["requests"]]
            return statuses

    def get_code_generators(self):
        with httpx.Client() as client:
            r = client.get(url=f"{self.url}/multiple-codegen-list")

            if r.status_code == 403:
                raise AuthorizationError(
                    f"Not authorized to access serviceX at {self.url}"
                )
            return r.json()

    async def get_datasets(
        self, did_finder=None, show_deleted=False
    ) -> List[CachedDataset]:
        headers = await self._get_authorization()
        params = {"did-finder": did_finder} if did_finder else {}
        if show_deleted:
            params["show-deleted"] = True

        async with ClientSession() as session:
            async with session.get(
                headers=headers, url=f"{self.url}/servicex/datasets", params=params
            ) as r:

                if r.status == 403:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status != 200:
                    msg = await _extract_message(r)
                    raise RuntimeError(f"Failed to get datasets: {r.status} - {msg}")

                result = await r.json()

            datasets = [CachedDataset(**d) for d in result["datasets"]]
            return datasets

    async def get_dataset(self, dataset_id=None) -> CachedDataset:
        headers = await self._get_authorization()
        path_template = "/servicex/datasets/{dataset_id}"
        url = self.url + path_template.format(dataset_id=dataset_id)
        async with ClientSession() as session:
            async with session.get(headers=headers, url=url) as r:

                if r.status == 403:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status == 404:
                    raise ValueError(f"Dataset {dataset_id} not found")
                elif r.status != 200:
                    msg = await _extract_message(r)
                    raise RuntimeError(f"Failed to get dataset {dataset_id} - {msg}")
                result = await r.json()

            dataset = CachedDataset(**result)
            return dataset

    async def delete_dataset(self, dataset_id=None) -> bool:
        headers = await self._get_authorization()
        path_template = "/servicex/datasets/{dataset_id}"
        url = self.url + path_template.format(dataset_id=dataset_id)

        async with ClientSession() as session:
            async with session.delete(headers=headers, url=url) as r:

                if r.status == 403:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status == 404:
                    raise ValueError(f"Dataset {dataset_id} not found")
                elif r.status != 200:
                    msg = await _extract_message(r)
                    raise RuntimeError(f"Failed to delete dataset {dataset_id} - {msg}")
                result = await r.json()
                return result["stale"]

    async def delete_transform(self, transform_id=None):
        headers = await self._get_authorization()
        path_template = f"/servicex/transformation/{transform_id}"
        url = self.url + path_template.format(transform_id=transform_id)

        async with ClientSession() as session:
            async with session.delete(headers=headers, url=url) as r:

                if r.status == 403:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status == 404:
                    raise ValueError(f"Transform {transform_id} not found")
                elif r.status != 200:
                    msg = await _extract_message(r)
                    raise RuntimeError(
                        f"Failed to delete transform {transform_id} - {msg}"
                    )

    async def cancel_transform(self, transform_id=None):
        headers = await self._get_authorization()
        path_template = f"/servicex/transformation/{transform_id}/cancel"
        url = self.url + path_template.format(transform_id=transform_id)

        async with ClientSession() as session:
            async with session.get(headers=headers, url=url) as r:

                if r.status == 403:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status == 404:
                    raise ValueError(f"Transform {transform_id} not found")
                elif r.status != 200:
                    msg = await _extract_message(r)
                    raise RuntimeError(
                        f"Failed to cancel transform {transform_id} - {msg}"
                    )

    async def submit_transform(self, transform_request: TransformRequest) -> str:
        headers = await self._get_authorization()
        retry_options = ExponentialRetry(attempts=3, start_timeout=30)
        async with RetryClient(retry_options=retry_options) as client:
            async with client.post(
                url=f"{self.url}/servicex/transformation",
                headers=headers,
                json=transform_request.model_dump(by_alias=True, exclude_none=True),
            ) as r:
                if r.status == 401:
                    raise AuthorizationError(
                        f"Not authorized to access serviceX at {self.url}"
                    )
                elif r.status == 400:
                    message = await _extract_message(r)
                    raise ValueError(f"Invalid transform request: {message}")
                elif r.status > 400:
                    error_message = await _extract_message(r)
                    raise RuntimeError(
                        "ServiceX WebAPI Error during transformation "
                        f"submission: {r.status} - {error_message}"
                    )
                else:
                    o = await r.json()
                    return o["request_id"]

    async def get_transform_status(self, request_id: str) -> TransformStatus:
        headers = await self._get_authorization()
        retry_options = ExponentialRetry(attempts=5, start_timeout=3)
        async with RetryClient(retry_options=retry_options) as client:
            try:
                async for attempt in AsyncRetrying(
                    retry=retry_if_not_exception_type(ValueError),
                    stop=stop_after_attempt(3),
                    wait=wait_fixed(3),
                    reraise=True,
                ):
                    with attempt:
                        async with client.get(
                            url=f"{self.url}/servicex/" f"transformation/{request_id}",
                            headers=headers,
                        ) as r:
                            if r.status == 401:
                                # perhaps we just ran out of auth validity the last time?
                                # refetch auth then raise an error for retry
                                headers = await self._get_authorization(True)
                                raise AuthorizationError(
                                    f"Not authorized to access serviceX at {self.url}"
                                )
                            if r.status == 404:
                                raise ValueError(f"Transform ID {request_id} not found")
                            elif r.status > 400:
                                error_message = await _extract_message(r)
                                raise RuntimeError(
                                    "ServiceX WebAPI Error during transformation: "
                                    f"{r.status} - {error_message}"
                                )
                            o = await r.json()
                            return TransformStatus(**o)
            except RuntimeError as e:
                raise RuntimeError(
                    "ServiceX WebAPI Error " f"while getting transform status: {e}"
                )



#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from unittest.mock import patch


def test_codegen_list(script_runner):
    with patch(
        "servicex.servicex_adapter.ServiceXAdapter.get_code_generators",
        return_value={"uproot": "http://uproot-codegen", "xaod": "http://xaod-codegen"},
    ):
        result = script_runner.run(
            ["servicex", "codegen", "list", "-c", "tests/example_config.yaml"]
        )
        assert result.returncode == 0
        assert (
            result.stdout
            == """{
  "uproot": "http://uproot-codegen",
  "xaod": "http://xaod-codegen"
}
"""
        )


def test_codegen_flush(script_runner):
    with patch("servicex.query_cache.QueryCache.delete_codegen_by_backend") as p:
        result = script_runner.run(
            [
                "servicex",
                "codegen",
                "flush",
                "-c",
                "tests/example_config.yaml",
                "-b",
                "localhost",
            ]
        )
        assert result.returncode == 0
        p.assert_called_once_with("localhost")



#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from pathlib import Path
from unittest.mock import patch, AsyncMock, Mock

import pytest

from servicex.models import (
    TransformStatus,
    ResultDestination,
    ResultFormat,
    Status,
    ResultFile,
)


@pytest.fixture
def transform_status_record() -> TransformStatus:
    """
    Pytest fixture generating a realistic TransformStatus instance for testing.

    Returns:
        TransformStatus: A populated TransformStatus object with sample data
    """

    # Default data that can be overridden
    base_data = {
        "request_id": "test-request-123",
        "did": "test-did-456",
        "title": "Test Transform Job",
        "selection": "(muon_pt > 20)",
        "tree-name": "mytree",
        "image": "servicex/transformer:latest",
        "result-destination": ResultDestination.object_store,
        "result-format": ResultFormat.parquet,
        "generated-code-cm": "generated-test-code",
        "status": Status.complete,
        "app-version": "1.2.3",
        "files": 10,
        "files-completed": 8,
        "files-failed": 1,
        "files-remaining": 1,
        "submit_time": datetime.now(),
        "finish_time": datetime.now(),
        "minio_endpoint": "minio.example.com",
        "minio_secured": True,
        "minio_access_key": "test-access-key",
        "minio_secret_key": "test-secret-key",
        "log_url": "https://logs.example.com/test-job",
    }

    return TransformStatus(**base_data)


@pytest.fixture
def result_files():
    return [
        ResultFile(filename="test_file", size=1024, extension="parquet"),
        ResultFile(filename="test_file2", size=2048, extension="parquet"),
    ]


def test_transforms_list(script_runner, transform_status_record):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        transform_status_record.status = Status.running
        mock_list_transforms = Mock(return_value=[transform_status_record])
        mock_servicex.return_value.get_transforms = mock_list_transforms
        result = script_runner.run(
            ["servicex", "transforms", "list", "-c", "tests/example_config.yaml"]
        )

        assert result.returncode == 0
        result_row = result.stdout.split("  ")
        assert len(result_row) == 4
        assert result_row[0].strip() == "test-request-123"
        assert result_row[1] == "Test Transform Job"
        assert result_row[2] == "Running"
        assert result_row[3].strip() == "8"

        mock_list_transforms.assert_called_once()
        mock_list_transforms.reset_mock()
        result = script_runner.run(
            [
                "servicex",
                "transforms",
                "list",
                "-c",
                "tests/example_config.yaml",
                "--complete",
            ]
        )
        assert result.returncode == 0
        assert len(result.stdout.strip()) == 0


@pytest.mark.parametrize(
    "transform_state, report_complete, report_running, expected",
    [
        # Scenario 1: No flags set (report all)
        (Status.complete, False, False, True),
        (Status.running, False, False, True),
        # Scenario 2: Complete records only
        (Status.complete, True, False, True),
        (Status.running, True, False, False),
        # Scenario 3: Running records only
        (Status.complete, False, True, False),
        (Status.running, False, True, True),
        # Scenario 4: Both flags set
        (Status.complete, True, True, True),
        (Status.running, True, True, True),
    ],
)
def test_transforms_list_filters(
    script_runner,
    transform_status_record,
    transform_state,
    report_complete,
    report_running,
    expected,
):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        transform_status_record.status = transform_state
        mock_list_transforms = Mock(return_value=[transform_status_record])
        mock_servicex.return_value.get_transforms = mock_list_transforms
        command_line = [
            "servicex",
            "transforms",
            "list",
            "-c",
            "tests/example_config.yaml",
        ]

        if report_complete:
            command_line.append("--complete")
        if report_running:
            command_line.append("--running")

        result = script_runner.run(command_line)

        assert result.returncode == 0
        assert (
            len(result.stdout.strip()) if expected else len(result.stdout.strip()) == 0
        )


def test_list_files(script_runner, transform_status_record, result_files):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        with patch("servicex.app.transforms.MinioAdapter") as mock_minio:
            mock_transform_status = AsyncMock(return_value=transform_status_record)
            mock_servicex.return_value.get_transform_status_async = (
                mock_transform_status
            )

            mock_minio_adapter = Mock()
            mock_minio_adapter.list_bucket = AsyncMock(return_value=result_files)
            mock_minio.for_transform = Mock(return_value=mock_minio_adapter)
            result = script_runner.run(
                [
                    "servicex",
                    "transforms",
                    "files",
                    "-c",
                    "tests/example_config.yaml",
                    "test-request-123",
                ]
            )
            assert result.returncode == 0
            result_rows = result.stdout.strip().split("\n")
            assert len(result_rows) == 2
            result_row = result_rows[1].split("  ")
            assert result_row[0].strip() == "test_file2"
            assert result_row[1] == "0.00"
            assert result_row[2] == "parquet"

            mock_transform_status.assert_called_once_with("test-request-123")
            mock_minio.for_transform.assert_called_once_with(transform_status_record)
            mock_minio_adapter.list_bucket.assert_called_once()


def test_download_files(script_runner, transform_status_record, result_files):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        with patch("servicex.app.transforms.MinioAdapter") as mock_minio:
            mock_transform_status = AsyncMock(return_value=transform_status_record)
            mock_servicex.return_value.get_transform_status_async = (
                mock_transform_status
            )

            mock_minio_adapter = Mock()
            mock_minio_adapter.list_bucket = AsyncMock(return_value=result_files)
            mock_minio_adapter.download_file = AsyncMock(
                return_value=Path("/tmp/test_file.parquet")
            )
            mock_minio.for_transform = Mock(return_value=mock_minio_adapter)
            result = script_runner.run(
                [
                    "servicex",
                    "transforms",
                    "download",
                    "-c",
                    "tests/example_config.yaml",
                    "test-request-123",
                ]
            )
            assert result.returncode == 0
            result_rows = result.stdout.strip().split("\n")
            assert len(result_rows) == 3
            assert result_rows[1] == "/tmp/test_file.parquet"

            mock_transform_status.assert_called_once_with("test-request-123")
            mock_minio.for_transform.assert_called_once_with(transform_status_record)
            mock_minio_adapter.list_bucket.assert_called_once()
            assert mock_minio_adapter.download_file.call_count == 2
            assert mock_minio_adapter.download_file.mock_calls[0].args[0] == "test_file"


def test_delete_transform(script_runner, transform_status_record):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        mock_delete_transform = AsyncMock(return_value=True)
        mock_servicex.return_value.delete_transform = mock_delete_transform

        mock_delete_local = Mock(return_value=True)
        mock_servicex.return_value.delete_local_transform = mock_delete_local

        result = script_runner.run(
            [
                "servicex",
                "transforms",
                "delete",
                "-c",
                "tests/example_config.yaml",
                "test-request-123",
            ]
        )
        assert result.returncode == 0
        assert result.stdout == "Transform test-request-123 deleted\n"
        mock_delete_transform.assert_called_once_with("test-request-123")


def test_cancel_transform(script_runner, transform_status_record):
    with patch("servicex.app.transforms.ServiceXClient") as mock_servicex:
        mock_cancel_transform = AsyncMock(return_value=True)
        mock_servicex.return_value.cancel_transform = mock_cancel_transform

        result = script_runner.run(
            [
                "servicex",
                "transforms",
                "cancel",
                "-c",
                "tests/example_config.yaml",
                "test-request-123",
            ]
        )
        assert result.returncode == 0
        assert result.stdout == "Transform test-request-123 cancelled\n"
        mock_cancel_transform.assert_called_once_with("test-request-123")



#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
from datetime import datetime
from unittest.mock import patch, AsyncMock

import pytest

from servicex.models import CachedDataset, DatasetFile


@pytest.fixture
def dataset():
    dataset_files = [
        DatasetFile(
            id=1,
            adler32="some_adler32_hash",
            file_size=1024,
            file_events=100,
            paths="/path/to/file1",
        ),
        DatasetFile(
            id=2,
            adler32="another_adler32_hash",
            file_size=2048,
            file_events=200,
            paths="/path/to/file2",
        ),
    ]

    cached_dataset = CachedDataset(
        id=42,
        name="test_dataset",
        did_finder="some_finder",
        n_files=2,
        size=3072,
        events=300,
        last_used=datetime.now(),
        last_updated=datetime.now(),
        lookup_status="completed",
        is_stale=False,
        files=dataset_files,
    )
    return cached_dataset


@pytest.mark.asyncio
def test_datasets_list(script_runner, dataset):
    with patch("servicex.app.datasets.ServiceXClient") as mock_servicex:
        mock_get_datasets = AsyncMock(return_value=[dataset])
        mock_servicex.return_value.get_datasets = mock_get_datasets

        result = script_runner.run(
            ["servicex", "datasets", "list", "-c", "tests/example_config.yaml"]
        )
        assert result.returncode == 0
        result_row = result.stdout.split("  ")
        assert len(result_row) == 7, f"Expected 7 elements, got {len(result_row)}"

        # Assert specific index values
        assert result_row[0].strip() == "42"
        assert result_row[1] == "test_dataset"
        assert result_row[2] == "2"
        assert result_row[3] == "0MB"
        assert result_row[4] == "completed"

        mock_get_datasets.assert_called_once_with(did_finder=None, show_deleted=False)

        mock_get_datasets.reset_mock()
        result = script_runner.run(
            [
                "servicex",
                "datasets",
                "list",
                "-c",
                "tests/example_config.yaml",
                "--did-finder",
                "some_finder",
                "--show-deleted",
            ]
        )
        assert result.returncode == 0
        mock_get_datasets.assert_called_once_with(
            did_finder="some_finder", show_deleted=True
        )


def test_dataset_get(script_runner, dataset):
    with patch("servicex.app.datasets.ServiceXClient") as mock_servicex:
        mock_get_dataset = AsyncMock(return_value=dataset)
        mock_servicex.return_value.get_dataset = mock_get_dataset

        result = script_runner.run(
            ["servicex", "datasets", "get", "42", "-c", "tests/example_config.yaml"]
        )
        assert result.returncode == 0
        mock_get_dataset.assert_called_once_with(42)

        # The output is a json document
        result_doc = json.loads(result.stdout)
        assert result_doc["dataset"]["id"] == 42
        assert len(result_doc["dataset"]["files"]) == 2


def test_dataset_delete(script_runner):
    with patch("servicex.app.datasets.ServiceXClient") as mock_servicex:
        mock_delete_dataset = AsyncMock(return_value=True)
        mock_servicex.return_value.delete_dataset = mock_delete_dataset

        result = script_runner.run(
            ["servicex", "datasets", "delete", "-c", "tests/example_config.yaml", "42"]
        )
        assert result.returncode == 0
        assert result.stdout == "Dataset 42 deleted\n"
        mock_delete_dataset.assert_called_once_with(42)

        mock_delete_dataset_not_found = AsyncMock(return_value=False)
        mock_servicex.return_value.delete_dataset = mock_delete_dataset_not_found
        result = script_runner.run(
            ["servicex", "datasets", "delete", "-c", "tests/example_config.yaml", "42"]
        )
        assert result.returncode == 1
        mock_delete_dataset.assert_called_once_with(42)
        assert result.stdout == "Dataset 42 not found\n"



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import ast
import copy
from typing import (
    Optional,
    Any,
    Type,
    TypeVar,
)
from qastle import python_ast_to_text_ast

from func_adl import EventDataset, find_EventDataset
from servicex.query_core import QueryStringGenerator
from abc import ABC

T = TypeVar("T")


class FuncADLQuery(QueryStringGenerator, EventDataset[T], ABC):
    r"""
    ServiceX Dataset class that uses func_adl query syntax.
    """

    # These are methods that are translated locally
    _execute_locally = ["ResultPandasDF", "ResultAwkwardArray"]
    default_codegen = None

    async def execute_result_async(
        self, a: ast.AST, title: Optional[str] = None
    ) -> Any:
        "Required override of EventDataset"

    def check_data_format_request(self, f_name: str):
        "Required override of EventDataset"

    def __init__(
        self,
        item_type: Type = Any,
    ):
        EventDataset.__init__(self, item_type=item_type)
        self.provided_qastle = None

    def set_provided_qastle(self, qastle: str):
        self.provided_qastle = qastle

    def generate_selection_string(self) -> str:
        if self.provided_qastle:
            return self.provided_qastle
        else:
            return self.generate_qastle(self.query_ast)

    def set_tree(self, tree_name: str) -> FuncADLQuery[T]:
        r"""Set the tree name for the query.
        Args:
            tree_name (str): Name of the tree to use for the query

        Returns:
            The Dataset with the tree appended to the first call object
        """

        clone = self.clone_with_new_ast(copy.deepcopy(self.query_ast), self._item_type)
        c = find_EventDataset(clone.query_ast)
        c.args.append(ast.Str(s="bogus.root"))
        c.args.append(ast.Str(s=tree_name))
        return clone

    def generate_qastle(self, a: ast.AST) -> str:
        r"""Generate the qastle from the ast of the query.

        1. The top level function is already marked as being "ok"
        1. If the top level function is something we have to process locally,
           then we strip it off.

        Args:
            a (ast.AST): The complete AST of the request.

        Returns:
            str: Qastle that should be sent to servicex
        """
        return python_ast_to_text_ast(a)

    def as_qastle(self):
        r"""
        Generate Qastle from this AST

        :returns:
            Qastle representation of the target's AST
        """
        return self.value()


class FuncADLQuery_Uproot(FuncADLQuery):
    yaml_tag = "!FuncADL_Uproot"
    default_codegen = "uproot"

    def __init__(
        self,
        item_type: Type = Any,
    ):
        super().__init__(item_type)
        self.tree_is_set = False

    def FromTree(self, tree_name):
        self.tree_is_set = True
        return self.set_tree(tree_name=tree_name)

    def generate_selection_string(self):
        if not self.tree_is_set:
            raise ValueError(
                "Uproot FuncADL query requires "
                "that you set a tree name with FromTree()"
            )
        return super().generate_selection_string()

    def set_provided_qastle(self, qastle_query: str):
        # we do not validate provided qastle, so we don't know if a tree name is specified.
        # assume user knows what they're doing
        self.tree_is_set = True
        super().set_provided_qastle(qastle_query)

    @classmethod
    def from_yaml(cls, _, node):
        import qastle
        import re

        # We have modified the funcADL spec a bit here. For uproot queries
        # we need to pick the optional tree specification up from the query
        # string and add it to the EventDataset constructor.
        from_tree_re = r'(?:FromTree\(["\']([^"\']+)["\']\)\.)?(.+)'
        tree_match = re.match(from_tree_re, node.value)

        if tree_match:
            query_string = (
                f"EventDataset('bogus.root', '{tree_match.group(1)}')."
                + tree_match.group(2)
            )
        else:
            query_string = "EventDataset('bogus.root', 'events')." + node.value

        qastle_query = qastle.python_ast_to_text_ast(
            qastle.insert_linq_nodes(ast.parse(query_string))
        )
        query = cls()
        query.set_provided_qastle(qastle_query)
        return query


class FuncADLQuery_ATLASr21(FuncADLQuery):
    yaml_tag = "!FuncADL_ATLASr21"
    default_codegen = "atlasr21"


class FuncADLQuery_ATLASr22(FuncADLQuery):
    yaml_tag = "!FuncADL_ATLASr22"
    default_codegen = "atlasr22"


class FuncADLQuery_ATLASxAOD(FuncADLQuery):
    yaml_tag = "!FuncADL_ATLASxAOD"
    default_codegen = "atlasxaod"


class FuncADLQuery_CMS(FuncADLQuery):
    yaml_tag = "!FuncADL_CMS"
    default_codegen = "cms"
